@@ -1,9 +1,11 @@
 import socket
 import time
 import logging
+import six
 
 from pymemcache.client.base import Client, PooledClient, _check_key
 from pymemcache.client.rendezvous import RendezvousHash
+from pymemcache.exceptions import MemcacheError
 
 logger = logging.getLogger(__name__)
 
@@ -12,10 +14,11 @@ class HashClient(object):
     """
     A client for communicating with a cluster of memcached servers
     """
+
     def __init__(
         self,
         servers,
-        hasher=None,
+        hasher=RendezvousHash,
         serializer=None,
         deserializer=None,
         connect_timeout=None,
@@ -30,6 +33,9 @@ class HashClient(object):
         dead_timeout=60,
         use_pooling=False,
         ignore_exc=False,
+        allow_unicode_keys=False,
+        default_noreply=True,
+        encoding='ascii'
     ):
         """
         Constructor.
@@ -50,14 +56,10 @@ class HashClient(object):
                                  attempts.
           dead_timeout (float): Time in seconds before attempting to add a node
                                 back in the pool.
+          encoding: optional str, controls data encoding (defaults to 'ascii').
 
         Further arguments are interpreted as for :py:class:`.Client`
         constructor.
-
-        The default ``hasher`` is using a pure python implementation that can
-        be significantly improved performance wise by switching to a C based
-        version. We recommend using ``python-clandestined`` if having a C
-        dependency is acceptable.
         """
         self.clients = {}
         self.retry_attempts = retry_attempts
@@ -66,12 +68,12 @@ class HashClient(object):
         self.use_pooling = use_pooling
         self.key_prefix = key_prefix
         self.ignore_exc = ignore_exc
+        self.allow_unicode_keys = allow_unicode_keys
         self._failed_clients = {}
         self._dead_clients = {}
         self._last_dead_check_time = time.time()
 
-        if hasher is None:
-            self.hasher = RendezvousHash()
+        self.hasher = hasher()
 
         self.default_kwargs = {
             'connect_timeout': connect_timeout,
@@ -81,6 +83,8 @@ class HashClient(object):
             'key_prefix': key_prefix,
             'serializer': serializer,
             'deserializer': deserializer,
+            'allow_unicode_keys': allow_unicode_keys,
+            'default_noreply': default_noreply,
         }
 
         if use_pooling is True:
@@ -91,6 +95,7 @@ class HashClient(object):
 
         for server, port in servers:
             self.add_server(server, port)
+        self.encoding = encoding
 
     def add_server(self, server, port):
         key = '%s:%s' % (server, port)
@@ -114,7 +119,7 @@ class HashClient(object):
         self.hasher.remove_node(key)
 
     def _get_client(self, key):
-        _check_key(key, self.key_prefix)
+        _check_key(key, self.allow_unicode_keys, self.key_prefix)
         if len(self._dead_clients) > 0:
             current_time = time.time()
             ldc = self._last_dead_check_time
@@ -131,6 +136,12 @@ class HashClient(object):
                         self._last_dead_check_time = current_time
 
         server = self.hasher.get_node(key)
+        # We've ran out of servers to try
+        if server is None:
+            if self.ignore_exc is True:
+                return
+            raise MemcacheError('All servers seem to be down right now')
+
         client = self.clients[server]
         return client
 
@@ -167,34 +178,7 @@ class HashClient(object):
         # Connecting to the server fail, we should enter
         # retry mode
         except socket.error:
-            # This client has never failed, lets mark it for failure
-            if (
-                    client.server not in self._failed_clients and
-                    self.retry_attempts > 0
-            ):
-                self._failed_clients[client.server] = {
-                    'failed_time': time.time(),
-                    'attempts': 0,
-                }
-            # We aren't allowing any retries, we should mark the server as
-            # dead immediately
-            elif (
-                client.server not in self._failed_clients and
-                self.retry_attempts < 0
-            ):
-                self._failed_clients[client.server] = {
-                    'failed_time': time.time(),
-                    'attempts': 0,
-                }
-                logger.debug("marking server as dead %s" % client.server)
-                self.remove_server(*client.server)
-            # This client has failed previously, we need to update the metadata
-            # to reflect that we have attempted it again
-            else:
-                failed_metadata = self._failed_clients[client.server]
-                failed_metadata['attempts'] += 1
-                failed_metadata['failed_time'] = time.time()
-                self._failed_clients[client.server] = failed_metadata
+            self._mark_failed_server(client.server)
 
             # if we haven't enabled ignore_exc, don't move on gracefully, just
             # raise the exception
@@ -202,7 +186,7 @@ class HashClient(object):
                 raise
 
             return default_val
-        except:
+        except Exception:
             # any exceptions that aren't socket.error we need to handle
             # gracefully as well
             if not self.ignore_exc:
@@ -210,8 +194,101 @@ class HashClient(object):
 
             return default_val
 
+    def _safely_run_set_many(self, client, values, *args, **kwargs):
+        failed = []
+        succeeded = []
+        try:
+            if client.server in self._failed_clients:
+                # This server is currently failing, lets check if it is in
+                # retry or marked as dead
+                failed_metadata = self._failed_clients[client.server]
+
+                # we haven't tried our max amount yet, if it has been enough
+                # time lets just retry using it
+                if failed_metadata['attempts'] < self.retry_attempts:
+                    failed_time = failed_metadata['failed_time']
+                    if time.time() - failed_time > self.retry_timeout:
+                        logger.debug(
+                            'retrying failed server: %s', client.server
+                        )
+                        succeeded, failed, err = self._set_many(
+                            client, values, *args, **kwargs)
+                        if err is not None:
+                            raise err
+                        # we were successful, lets remove it from the failed
+                        # clients
+                        self._failed_clients.pop(client.server)
+                        return failed
+                    return values.keys()
+                else:
+                    # We've reached our max retry attempts, we need to mark
+                    # the sever as dead
+                    logger.debug('marking server as dead: %s', client.server)
+                    self.remove_server(*client.server)
+
+            succeeded, failed, err = self._set_many(
+                client, values, *args, **kwargs
+            )
+            if err is not None:
+                raise err
+
+            return failed
+
+        # Connecting to the server fail, we should enter
+        # retry mode
+        except socket.error:
+            self._mark_failed_server(client.server)
+
+            # if we haven't enabled ignore_exc, don't move on gracefully, just
+            # raise the exception
+            if not self.ignore_exc:
+                raise
+
+            return list(set(values.keys()) - set(succeeded))
+        except Exception:
+            # any exceptions that aren't socket.error we need to handle
+            # gracefully as well
+            if not self.ignore_exc:
+                raise
+
+            return list(set(values.keys()) - set(succeeded))
+
+    def _mark_failed_server(self, server):
+        # This client has never failed, lets mark it for failure
+        if (
+                server not in self._failed_clients and
+                self.retry_attempts > 0
+        ):
+            self._failed_clients[server] = {
+                'failed_time': time.time(),
+                'attempts': 0,
+            }
+        # We aren't allowing any retries, we should mark the server as
+        # dead immediately
+        elif (
+            server not in self._failed_clients and
+            self.retry_attempts <= 0
+        ):
+            self._failed_clients[server] = {
+                'failed_time': time.time(),
+                'attempts': 0,
+            }
+            logger.debug("marking server as dead %s", server)
+            self.remove_server(*server)
+        # This client has failed previously, we need to update the metadata
+        # to reflect that we have attempted it again
+        else:
+            failed_metadata = self._failed_clients[server]
+            failed_metadata['attempts'] += 1
+            failed_metadata['failed_time'] = time.time()
+            self._failed_clients[server] = failed_metadata
+
     def _run_cmd(self, cmd, key, default_val, *args, **kwargs):
         client = self._get_client(key)
+
+        if client is None:
+            return default_val
+
         func = getattr(client, cmd)
         args = list(args)
         args.insert(0, key)
@@ -219,6 +296,22 @@ class HashClient(object):
             client, func, default_val, *args, **kwargs
         )
 
+    def _set_many(self, client, values, *args, **kwargs):
+        failed = []
+        succeeded = []
+
+        try:
+            for key, value in six.iteritems(values):
+                result = client.set(key, value, *args, **kwargs)
+                if result:
+                    succeeded.append(key)
+                else:
+                    failed.append(key)
+        except Exception as e:
+            return succeeded, failed, e
+
+        return succeeded, failed, None
+
     def set(self, key, *args, **kwargs):
         return self._run_cmd('set', key, False, *args, **kwargs)
 
@@ -233,47 +326,60 @@ class HashClient(object):
 
     def set_many(self, values, *args, **kwargs):
         client_batches = {}
-        for key, value in values.items():
+        failed = []
+
+        for key, value in six.iteritems(values):
             client = self._get_client(key)
+
+            if client is None:
+                failed.append(key)
+                continue
+
             if client.server not in client_batches:
                 client_batches[client.server] = {}
 
             client_batches[client.server][key] = value
 
-        end = []
-
         for server, values in client_batches.items():
             client = self.clients['%s:%s' % server]
-            new_args = list(args)
-            new_args.insert(0, values)
-            result = self._safely_run_func(
-                client,
-                client.set_many, False, *new_args, **kwargs
-            )
-            end.append(result)
 
-        return all(end)
+            failed += self._safely_run_set_many(
+                client, values, *args, **kwargs
+            )
+
+        return failed
 
     set_multi = set_many
 
-    def get_many(self, keys, *args, **kwargs):
+    def get_many(self, keys, gets=False, *args, **kwargs):
         client_batches = {}
+        end = {}
+
         for key in keys:
             client = self._get_client(key)
+
+            if client is None:
+                end[key] = False
+                continue
+
             if client.server not in client_batches:
                 client_batches[client.server] = []
 
             client_batches[client.server].append(key)
 
-        end = {}
-
         for server, keys in client_batches.items():
             client = self.clients['%s:%s' % server]
             new_args = list(args)
             new_args.insert(0, keys)
+
+            if gets:
+                get_func = client.gets_many
+            else:
+                get_func = client.get_many
+
             result = self._safely_run_func(
                 client,
-                client.get_many, {}, *new_args, **kwargs
+                get_func, {}, *new_args, **kwargs
             )
             end.update(result)
 
@@ -284,6 +390,11 @@ class HashClient(object):
     def gets(self, key, *args, **kwargs):
         return self._run_cmd('gets', key, None, *args, **kwargs)
 
+    def gets_many(self, keys, *args, **kwargs):
+        return self.get_many(keys, gets=True, *args, **kwargs)
+
+    gets_multi = gets_many
+
     def add(self, key, *args, **kwargs):
         return self._run_cmd('add', key, False, *args, **kwargs)
 
