@@ -25,12 +25,16 @@ class MockMemcacheClient(object):
                  connect_timeout=None,
                  timeout=None,
                  no_delay=False,
-                 ignore_exc=False):
+                 ignore_exc=False,
+                 default_noreply=True,
+                 allow_unicode_keys=False,
+                 encoding='ascii'):
 
         self._contents = {}
 
         self.serializer = serializer
         self.deserializer = deserializer
+        self.allow_unicode_keys = allow_unicode_keys
 
         # Unused, but present for interface compatibility
         self.server = server
@@ -38,21 +42,29 @@ class MockMemcacheClient(object):
         self.timeout = timeout
         self.no_delay = no_delay
         self.ignore_exc = ignore_exc
+        self.encoding = encoding
 
-    def get(self, key):
-        if isinstance(key, six.text_type):
-            raise MemcacheIllegalInputError(key)
+    def get(self, key, default=None):
+        if not self.allow_unicode_keys:
+            if isinstance(key, six.string_types):
+                try:
+                    if isinstance(key, bytes):
+                        key = key.decode().encode('ascii')
+                    else:
+                        key = key.encode('ascii')
+                except (UnicodeEncodeError, UnicodeDecodeError):
+                    raise MemcacheIllegalInputError
 
         if key not in self._contents:
-            return None
+            return default
 
-        expire, value, was_serialized = self._contents[key]
+        expire, value, flags = self._contents[key]
         if expire and expire < time.time():
             del self._contents[key]
-            return None
+            return default
 
         if self.deserializer:
-            return self.deserializer(key, value, 2 if was_serialized else 1)
+            return self.deserializer(key, value, flags)
         return value
 
     def get_many(self, keys):
@@ -65,26 +77,37 @@ class MockMemcacheClient(object):
 
     get_multi = get_many
 
-    def set(self, key, value, expire=0, noreply=True):
-        if isinstance(key, six.text_type):
-            raise MemcacheIllegalInputError(key)
-        if isinstance(value, six.text_type):
-            raise MemcacheIllegalInputError(value)
+    def set(self, key, value, expire=0, noreply=True, flags=0):
+        if not self.allow_unicode_keys:
+            if isinstance(key, six.string_types):
+                try:
+                    if isinstance(key, bytes):
+                        key = key.decode().encode()
+                    else:
+                        key = key.encode(self.encoding)
+                except (UnicodeEncodeError, UnicodeDecodeError):
+                    raise MemcacheIllegalInputError
+        if (isinstance(value, six.string_types) and
+                not isinstance(value, six.binary_type)):
+            try:
+                value = value.encode(self.encoding)
+            except (UnicodeEncodeError, UnicodeDecodeError):
+                raise MemcacheIllegalInputError
 
-        was_serialized = False
+        flags = 0
         if self.serializer:
-            value = self.serializer(key, value)
+            value, flags = self.serializer(key, value)
 
         if expire:
             expire += time.time()
 
-        self._contents[key] = expire, value, was_serialized
+        self._contents[key] = expire, value, flags
         return True
 
     def set_many(self, values, expire=None, noreply=True):
         for key, value in six.iteritems(values):
             self.set(key, value, expire, noreply)
-        return True
+        return []
 
     set_multi = set_many
 
